#ifndef ALIGNER_HH
#define ALIGNER_HH

#include "base.hh"
#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"

/*
Paramater for Aligner

*/

class AlignerParams
{  
protected:
  const int match_score;
  const int mismatch_score;
  const int indel_score;

  const int tau_factor; // contribution of sequence similarity in an arc match (in percent)
  
  const bool no_lonely_pair; // no lonely pairs option
  
  const bool STRUCT_LOCAL; // allow exclusions for maximizing alignment of connected substructures 
  const bool SEQU_LOCAL; // maximize alignment of subsequences
  
  const bool DO_TRACE;
    
  const int exclusion_cost; // Cost/similarity contribution per exclusion
  
public:
  AlignerParams(int _match_score,int _mismatch_score,int _indel_score, int _tau_factor,
		bool _no_lonely_pair, bool _STRUCT_LOCAL, bool _SEQU_LOCAL):
    match_score(_match_score),
    mismatch_score(_mismatch_score),
    indel_score(_indel_score),
    tau_factor(_tau_factor),
    no_lonely_pair(_no_lonely_pair),
    STRUCT_LOCAL(_STRUCT_LOCAL),
    SEQU_LOCAL(_SEQU_LOCAL),
    DO_TRACE(true),
    exclusion_cost(0)
  {}
  
};



/* ============================================================ 
   class Aligner

   Compute alignment similarity score and optimal alignment
   
   Methods:
   - Aligner
   initialize with two sequences and their basepairs (including scores)
   - sigma
   similartiy for bases
   - tau 
   sequence contribution for arc-matches
   - ...
============================================================ */

#define FOR_IJ_IN_ARCMATCH \
  for (int i=al+1; i<ar; i++) \
    for (int j=bl+1; j<br; j++)

class Aligner : AlignerParams{
  sequence_t seqA;
  const Bps &bpsA;
  sequence_t seqB;
  const Bps &bpsB;
  
  typedef std::vector<std::vector<score_t> > score_matrix_t;
  
  score_matrix_t D;
  std::vector<score_matrix_t> Ms;
  
  Alignment alignment;
  
  enum {E_NO_NO, E_X_NO, E_NO_X, E_X_X,
	E_OP_NO, E_NO_OP, E_OP_X, E_X_OP};
  
  int max_i,max_j; // computed by align_top_level
  
  
  int sigma_(int i, int j);
  void precompute_sigma();  
  std::vector<std::vector<int> > sigma_tab;
  
  int sigma(int i, int j) {return sigma_tab[i][j];}

  int tau(int i, int j, int k, int l) {
    return (tau_factor)?( ((sigma(i,j)+sigma(k,l)) * tau_factor) / 100 ):0;
  }


  // initialize first column and row
  void init_state(score_matrix_t &M, int al, int ar, int bl, int br, 
		  score_t indelA, score_t exclusion_costA,
		  score_t indelB, score_t exclusion_costB);
  
  // recursion cases that handle everything but exclusions
  // (in the LSSA-paper this function was called NoEx
  
  score_t align_noex(score_matrix_t &M,int al,int bl,int i, int j);
     
  // align, where al,ar,bl,br is an arc-match
  // in structure local alignment, this allows to introduce exclusions
    //
  void align_in_arcmatch(int al,int ar,int bl,int br,
			 bool allow_exclusion);
  
  // align the top-level in a sequence local alignment
  // and return the maximal score
  int align_top_level();
  
  void trace_in_arcmatch(int state,int al,int i,int bl,int j,bool top_level);
  
  void trace_noex(int state,int al,int i,int bl,int j,bool top_level);

public:  
  Aligner(sequence_t _seqA, const Bps &_bpsA,
	  sequence_t _seqB, const Bps &_bpsB,
	  const AlignerParams &_ap	    
	  );
  
  Alignment const & get_alignment() const {return alignment;} 
  
  score_t align();
  
  void trace();
};


#endif
