#include <string>
#include <fstream>
#include <map>

#include <iostream>
#include <fstream>
#include <sstream>


#include "basepairs.hh"
#include "math.h"

using namespace std;

/*! compare right ends of arcs (for sorting) */
bool operator < (const LeftAdjEntry &e1,const LeftAdjEntry &e2) {
    return e1.right<e2.right;
}

/*! compare left ends of arcs (for sorting) */
bool operator < (const RightAdjEntry &e1,const RightAdjEntry &e2) {
    return e1.left>e2.left;
}

/*! resize data structures that allow fast acces to an arc arcs by
   its left end, its right end, or its left and right end
*/

void BasePairs::resize(uint s) {
  s_=s;
  
  left_.resize(s);
  right_.resize(s);
  
#ifndef SPACE_OPT

  arcs_.resize(s);
  for (uint i=0; i<s; i++)
    arcs_[i].resize(s);
  
  for (uint i=0; i<s; i++)
    for (uint j=0; j<s; j++)
      arcs_[i][j]=-1;
  
  // all arc probabilities
  arc_probs_.resize(s);
  for (uint i=0; i<s; i++)
    arc_probs_[i].resize(s);
  
  for (uint i=0; i<s; i++)
    for (uint j=0; j<s; j++)
      arc_probs_[i][j]=0;

#endif

}

double BasePairs::prob_min() const {
    return min_prob;
}

double BasePairs::prob_exp() const {
    if (opt_exp_prob) {
	return  exp_prob;
    } else {
	assert(len_ != 0);
	return prob_exp_f(len_);
    }
}

int BasePairs::probToWeight(double p) const {
    double pe=prob_exp();
    
    return
	(int)round(structure_weight*
		   (1-log(p)/log(pe))
	    );
    
    /*
      Score soll 0 sein für p=pe und structure_weight für p=1
      
      ==> Normalisierung log(p/pe) / log(1/pe) 
      //                  [=1-log(p)/log(pe)]

    */
    
}


void BasePairs::insert(int i, int j, score_t w) {
    if (i>j) swap(i,j);
    
    int idx=arc_vec_.size();

    arc_vec_.push_back(Arc(idx,w));

    left_[i].push_back(LeftAdjEntry(idx,j,w));
    right_[j].push_back(RightAdjEntry(idx,i,w));

#ifdef SPACE_OPT    
    arcs_.set(i,j,idx);
#else
    arcs_[i][j]=idx;
#endif
}

void BasePairs::sortAdjLists() {
    for (uint i=0; i<left_.size(); ++i)
	sort(left_[i].begin(),left_[i].end());
    for (uint i=0; i<right_.size(); ++i)
	sort(right_[i].begin(),right_[i].end());
}

string seqname_from_filename(const string &s) {
    int i;
    int j;
    for (i=s.length()-1; i>=0 && s[i]!='/'; i--)
	;

    for (j=i+1; j<s.length() && s[j]!='.'; j++)
	;

    string name=s.substr(i+1,j-i-1);

    if (name.length()>3 && name.substr(name.length()-3,3) == "_dp")
	name=name.substr(0,name.length()-3);
    
    return name;
}

void BasePairs::read(char *filename, sequence_t &seq) {
	
    ifstream in(filename);
    if (! in.good()) {
		cerr << "Cannot read "<<filename<<endl;
		exit(-1);
    }
    
    string s;
    // read first string and decide about file-format
    in >> s;
    in.close();
    if (s == "%!PS-Adobe-3.0") {
		// try reading as dot.ps file (as generated by RNAfold)
		readPS(filename,seq);
    } else {
		// try reading as PP-file (proprietary format, which is easy to read and contains pair probs)
		readPP(filename,seq);
    }
    
}

void BasePairs::readPS(char *filename, sequence_t &seq) {
    ifstream in(filename);

    string s;
    while (in >> s && s!="/sequence")
	;
    in >> s; in >> s;

    string seqstr="";
    while (in >> s && s!=")") {    
		s = s.substr(0,s.size()-1); // chop of last character
		// cout << s <<endl;
		seqstr+=s;
    }
    
    seq.append_row(seqname_from_filename(filename),seqstr);
    
    len_ = seq.length();
    
    resize(len_+1);
    
    char buf[1024];
    while (in.getline(buf,1024)) {
		if (strcmp(buf+strlen(buf)-4,"ubox")==0) {
	
			istringstream ss(buf);
			int i,j;
			double p;
			ss >> i >> j >> p;
			
			p*=p;

			assert(i<j);
			
#ifdef SPACE_OPT
			arc_probs_.set(i,j,p);
#else
			arc_probs_[i][j]=p;
#endif

			if (p>=min_prob) {
				insert(i,j, probToWeight(p));
			}
		}
    }
	
    sortAdjLists();
}

void BasePairs::readPP(char *filename, sequence_t &seq) {
    ifstream in(filename);
    
    string name;
    string seqstr;
    
    map<string,string> seq_map;
    
    const int BUF_SIZE=4096;
    char buf[BUF_SIZE];
    while (in.getline(buf,BUF_SIZE) && strcmp(buf,"#")!=0 ) {
	if (strlen(buf)>0 && buf[0]!=' ') {
	    istringstream in(buf);
	    in >> name >> seqstr;
	    seq_map[name]+=seqstr;
	}
    }
    
    for (map<string,string>::iterator it=seq_map.begin(); it!=seq_map.end(); ++it) {
	seq.append_row(it->first,it->second);
    }

    int i,j;
    double p;
    
    len_=seq.length();

    resize(len_+1);

    while( in >>i>>j>>p ) {
	assert(i<j);
	
#ifdef SPACE_OPT
	arc_probs_.set(i,j,p);
#else
	arc_probs_[i][j]=p;
#endif
		
	if (p>=min_prob) {
	    int weight = probToWeight(p);
	    insert(i,j, weight);
	}
    }
    sortAdjLists();
}

