/**********************************************************************
 *
 * Locarna: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will@informatik.uni-freiburg.de> 2005, 2006
 *
 **********************************************************************/


#include <iostream>
#include <fstream>
// #include <sstream>
#include <vector>

#include <math.h>

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"
#include "aligner.hh"


using namespace std;

const string VERSION_STRING="LocARNA 0.975 --- pairwise Local Alignment of RNAs";

// ------------------------------------------------------------
// Parameter

double min_prob; // only pairs with a probability of at least min_prob are taken into account

int match_score;
int mismatch_score;
int indel_score;

int structure_weight;

int tau_factor; // contribution of sequence similarity in an arc match (in percent)
               
bool no_lonely_pair; // no lonely pairs option

bool STRUCT_LOCAL; // allow exclusions for maximizing alignment of connected substructures 
bool SEQU_LOCAL; // maximize alignment of subsequences

const bool DO_TRACE=true;


const int max_diff = -1; // maximal difference between two arc ends, -1 is off
// MAX_DIFF NOT IMPLEMENTED:
//  plan: do not match arcs where
//   1. for global (bl-al)>max_diff || (br-ar)>max_diff
//   2. for local (ar-al)-(br-bl)>max_diff
//   except when there is no additional computation of M matrices necessary.

const int exclusion_cost = 0; // Cost/similarity contribution per exclusion
                              // set to zero for unrestricted structure locality

double prob_exp_f(int seqlen) {return 1.0/(2*seqlen);} // expected probability of a base pair (null-model)

double exp_prob;
bool opt_exp_prob;

int output_width;

// ------------------------------------------------------------
// File arguments
char *bpsfile1;
char *bpsfile2;

char *clustal_out;
bool opt_clustal_out;

char *pp_out;
bool opt_pp_out;

// ------------------------------------------------------------
//
// Options
//
#include "options.hh"

bool opt_help;
bool opt_version;
bool opt_verbose;

option_def my_options[] = {
  {"help",'h',&opt_help,O_NO_ARG,0,0,0,"This help"},
  {"version",'V',&opt_version,O_NO_ARG,0,0,0,"Version info"},
  {"verbose",'v',&opt_verbose,O_NO_ARG,0,0,0,"Verbose"},
  {"match",'m',0,O_ARG_INT,&match_score,"50","s","Match score"},
  {"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","s","Mismatch score"},
  {"indel",'i',0,O_ARG_INT,&indel_score,"-200","s","Indel score"},
  {"struct-weight",'s',0,O_ARG_INT,&structure_weight,"100","s","Maximal weight of 1/2 arc-match"},
  {"tau",'t',0,O_ARG_INT,&tau_factor,"0","f","Tau factor in percent"},
  {"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"1e-2","f","Minimal probability"},
  {"exp-prob",'e',&opt_exp_prob,O_ARG_DOUBLE,&exp_prob,0,"f","Expected probability"},
  {"noLP",'l',&no_lonely_pair,O_NO_ARG,0,0,0,"No lonely pair"},
  {"struct-local",0,0,O_ARG_BOOL,&STRUCT_LOCAL,"false","bool","Structure local"},
  {"sequ-local",0,0,O_ARG_BOOL,&SEQU_LOCAL,"true","bool","Sequence local"},
  {"clustal",0,&opt_clustal_out,O_ARG_STRING,&clustal_out,0,"file","Clustal output"},
  {"pp",0,&opt_pp_out,O_ARG_STRING,&pp_out,0,"file","PP output"},
  {"width",'w',0,O_ARG_INT,&output_width,"60","columns","Output width"},
  {0,0,0,O_ARG_STRING,&bpsfile1,0,"bps-file 1","Basepairs input file 1"},
  {0,0,0,O_ARG_STRING,&bpsfile2,0,"bps-file 2","Basepairs input file 2"},
  {0,0,0,0,0,0,0,0}
};


// ------------------------------------------------------------


/* ------------------------------------------------------------
   Experimental feature: adapt score to sequence identity
   ------------------------------------------------------------ */

double seq_identity(sequence_t a, sequence_t b) {
    const size_t n=a.length();
    const size_t m=b.length();

    vector<vector<double> > M;

    M.resize(n+1);
    for (uint i=0; i<=n; i++) M[i].resize(m+1);
        
    for (uint i=0; i<=n; i++) M[i][0]=0;
    for (uint j=0; j<=m; j++) M[0][j]=0;

    for (uint i=1; i<=n; i++)
	for (uint j=1; j<=m; j++)
	    M[i][j] = max(M[i-1][j-1] + a[i].prob_equals(b[j]),
			  max(M[i-1][j],
			       M[i][j-1]));
			      
    return 2.0*M[n][m]/(double)(n+m);
}



/* ------------------------------------------------------------ */



// ------------------------------------------------------------
// MAIN

int
main(int argc, char **argv) {
    
    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	cout << VERSION_STRING<<endl<<endl;
	print_help(argv[0],my_options);
	exit(0);
    }
    
    if (opt_version) {
	cout << VERSION_STRING<<endl<<endl;
	exit(0);
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }
    
    if (opt_verbose)
	print_options(my_options);
    
    sequence_t seqA;
    sequence_t seqB;

    BasePairsParams bpp(structure_weight, min_prob, 
			exp_prob, opt_exp_prob);

    BasePairs bpsA(bpp);
    BasePairs bpsB(bpp);
    
    bpsA.read(bpsfile1,seqA);
    bpsB.read(bpsfile2,seqB);
    
    if (opt_verbose) {
	cout << "Sequence A: "<<endl;
	seqA.write(cout);
	cout<<" (Length:"<< seqA.length()<<", Basepairs:"<<bpsA.num_bps() << ")" <<endl;
	
	cout << "Sequence B: "<<endl;
	seqB.write(cout);
	cout<<" (Length:"<< seqB.length()<<", Basepairs:"<<bpsB.num_bps() << ")" <<endl;

	cout << "Base Identity: "<<(seq_identity(seqA,seqB)*100)<<endl; 
    }
    
    AlignerParams 
      ap(match_score, mismatch_score, indel_score,
	 tau_factor, no_lonely_pair, STRUCT_LOCAL, SEQU_LOCAL);

    Aligner aligner(seqA,bpsA,seqB,bpsB,ap);
    
    score_t score = aligner.align();
    
    cout << "Score: "<<score<<endl;
    
    if (DO_TRACE) {
      aligner.trace();
      aligner.get_alignment().write(cout, output_width);
      cout<<endl;
      
      if (opt_clustal_out) {
	ofstream out(clustal_out);
	if (out) {
	  aligner.get_alignment().
	    write_clustal(out, output_width, score);
	} else {
	  cerr << "Cannot write to "<<clustal_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
	if (opt_pp_out) {
	    ofstream out(pp_out);
	    if (out) {
		aligner.get_alignment().
		  write_pp(out,bpsA,bpsB,output_width);
	    } else {
		cerr << "Cannot write to "<<pp_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}

    }
    exit(0);
}
